import os

band1 = os.path.join(os.path.dirname(__file__), 'lsat7_2000_10.tif')
band2 = os.path.join(os.path.dirname(__file__), 'lsat7_2000_20.tif')
band3 = os.path.join(os.path.dirname(__file__), 'lsat7_2000_30.tif')
band4 = os.path.join(os.path.dirname(__file__), 'lsat7_2000_40.tif')
band5 = os.path.join(os.path.dirname(__file__), 'lsat7_2000_50.tif')
band7 = os.path.join(os.path.dirname(__file__), 'lsat7_2000_70.tif')
strata = os.path.join(os.path.dirname(__file__), 'strata.tif')
multiband = os.path.join(os.path.dirname(__file__), 'landsat_multiband.tif')
labelled_pixels = os.path.join(os.path.dirname(__file__), 'landsat96_labelled_pixels.tif')
points = os.path.join(os.path.dirname(__file__), 'landsat96_points.shp')
polygons = os.path.join(os.path.dirname(__file__), 'landsat96_polygons.shp')
extracted_pixels = os.path.join(os.path.dirname(__file__), 'extracted_pixels.txt')
